use std::collections::HashSet;
use std::marker::PhantomData;
use std::mem::size_of;
use std::ops::Range;

use zerocopy::{AsBytes, FromBytes, FromZeroes, Unaligned};

use crate::mapped::Slabable;
use crate::ReaderTrait;
use crate::{transaction, FilePointer, TransactionHandle, U64};

#[derive(Clone, Copy, FromBytes, FromZeroes, AsBytes, Unaligned)]
#[repr(packed)]
pub struct Queue<T: FromBytes + FromZeroes + AsBytes + Unaligned + Clone + Copy> {
    pub(crate) head: FilePointer<QueueElement<T>>,
    pub(crate) length: U64,
    _phantom: PhantomData<T>,
}

impl<T: FromBytes + FromZeroes + AsBytes + Unaligned + Clone + Copy> Slabable for Queue<T> {
    fn slabs(slabs: &mut HashSet<u32>) {
        slabs.insert(size_of::<Queue<T>>() as u32);
        slabs.insert(size_of::<QueueElement<T>>() as u32);
    }
}

// TODO: make this return the bare Queue to avoid additional allocations
impl<T: FromBytes + FromZeroes + AsBytes + Unaligned + Clone + Copy> Queue<T> {
    pub fn new<R>(transaction: &mut TransactionHandle<R>) -> FilePointer<Queue<T>> {
        let (queue, data) = transaction.allocate();
        *data = Self {
            head: FilePointer::null(),
            length: 0.into(),
            _phantom: PhantomData,
        };
        queue
    }
}

impl<T: FromBytes + FromZeroes + AsBytes + Unaligned + Clone + Copy> FilePointer<Queue<T>> {
    pub fn enqueue<R>(self, transaction: &mut TransactionHandle<R>, item: T) -> Self {
        self.enqueue_many(transaction, &[item])
    }

    pub fn dequeue<R>(self, transaction: &mut TransactionHandle<R>) -> Option<(Self, T)> {
        let mut res = None;
        self.dequeue_many_inner(transaction, 1, |t| assert!(res.replace(t).is_none()))
            .map(|ptr| (ptr, res.unwrap()))
    }

    pub fn enqueue_many<R>(self, transaction: &mut TransactionHandle<R>, items: &[T]) -> Self {
        if items.is_empty() {
            return self;
        }

        let old = *transaction.read(self);

        let mut head = old.head;

        for item in items {
            let (element, data) = transaction.allocate();
            *data = QueueElement {
                next: head,
                data: *item,
                _phantom: PhantomData,
            };

            head = element;
        }

        let (queue, data) = transaction.modify(self);
        *data = Queue {
            head,
            length: (old
                .length
                .get()
                .checked_add(items.len().try_into().unwrap())
                .unwrap())
            .into(),
            _phantom: PhantomData,
        };
        queue
    }

    pub fn dequeue_many<R>(
        self,
        transaction: &mut TransactionHandle<R>,
        n: u64,
    ) -> Option<(Self, Vec<T>)> {
        let mut res = Vec::with_capacity(n.try_into().unwrap());
        self.dequeue_many_inner(transaction, n, |t| res.push(t))
            .map(|ptr| {
                res.reverse();
                (ptr, res)
            })
    }

    // NOTE: calls f with the elements in reverse order.
    fn dequeue_many_inner<R>(
        self,
        transaction: &mut TransactionHandle<R>,
        n: u64,
        mut f: impl FnMut(T),
    ) -> Option<Self> {
        if n == 0 {
            return Some(self);
        }

        let old = *transaction.read(self);

        if old.length.get() < n {
            return None;
        }

        let (queue, data) = transaction.modify(self);

        *data = Queue {
            head: old.head,
            length: (old.length.get().checked_sub(n).unwrap()).into(),
            _phantom: PhantomData,
        };

        let mut ptr = data.head;

        let skip_n = old.length.get().checked_sub(n).unwrap();

        // skip the head
        for i in 0..skip_n {
            ptr = *transaction.read(field_ptr!(ptr, QueueElement<T>, next));
        }

        // remove the tail
        for i in 0..n {
            let element = *transaction.read(ptr);
            f(element.data);
            transaction.free(ptr);
            ptr = element.next;
        }

        Some(queue)
    }

    pub fn dequeue_many_back<R>(
        self,
        transaction: &mut TransactionHandle<R>,
        n: u64,
    ) -> Option<(Self, Vec<T>)> {
        let mut res = Vec::with_capacity(n.try_into().unwrap());
        self.dequeue_many_inner(transaction, n, |t| res.push(t))
            .map(|ptr| {
                res.reverse();
                (ptr, res)
            })
    }

    // NOTE: calls f with the elements in reverse order.
    fn dequeue_many_back_inner<R>(
        self,
        transaction: &mut TransactionHandle<R>,
        n: u64,
        mut f: impl FnMut(T),
    ) -> Option<Self> {
        if n == 0 {
            return Some(self);
        }

        let old = *transaction.read(self);

        if old.length.get() < n {
            return None;
        }

        let mut ptr = old.head;

        for i in 0..n {
            let element = *transaction.read(ptr);
            f(element.data);
            transaction.free(ptr);
            ptr = element.next;
        }

        let (queue, data) = transaction.modify(self);

        *data = Queue {
            head: ptr,
            length: (old.length.get().checked_sub(n).unwrap()).into(),
            _phantom: PhantomData,
        };

        Some(queue)
    }

    pub fn is_empty(self, reader: &impl ReaderTrait) -> bool {
        self.length(reader) == 0
    }

    // TODO: use get(0) instead
    pub fn next(self, reader: &impl ReaderTrait) -> Option<T> {
        let this = reader.read(self);

        if this.length.get() == 0 {
            return None;
        }

        let mut next = this.head;

        for _ in 1..this.length.get() {
            // println!("free [{i}] {ptr:?}");

            next = *reader.read(field_ptr!(next, QueueElement<T>, next));
        }

        Some(*reader.read(field_ptr!(next, QueueElement<T>, data)))
    }

    pub fn get_range(self, reader: &impl ReaderTrait, range: Range<u64>) -> Option<Vec<T>> {
        let mut res = Vec::with_capacity(range.clone().count());

        self.get_range_inner(reader, range, |element| {
            res.push(element);
        });

        res.reverse();
        Some(res)
    }

    // 0 is the next item to be dequeued
    pub fn get_range_inner(
        self,
        reader: &impl ReaderTrait,
        range: Range<u64>,
        mut f: impl FnMut(T),
    ) -> Option<()> {
        if range.is_empty() {
            return Some(());
        }

        let n = range.clone().count();

        let this = reader.read(self);

        if range.end > this.length.get() {
            return None;
        }

        let len = this.length.get();
        let start = len.checked_sub(range.end).unwrap();
        let end = len.checked_sub(range.start).unwrap();

        let range = start..end;

        let mut ptr = this.head;

        // skip the elements before the start of the list
        for i in 0..range.start {
            ptr = *reader.read(field_ptr!(ptr, QueueElement<T>, next));
        }

        for _ in 0..n {
            let element = *reader.read(ptr);

            f(element.data);

            ptr = element.next;
        }

        Some(())
    }

    // 0 ist the next item to be returned
    pub fn get(self, reader: &impl ReaderTrait, index: u64) -> Option<T> {
        let mut res = None;
        self.get_range_inner(reader, index..index + 1, |element| {
            assert!(res.replace(element).is_none());
        })?;
        res
    }

    pub fn last(self, reader: &impl ReaderTrait) -> Option<T> {
        let length = self.length(reader);

        if length == 0 {
            None
        } else {
            self.get(reader, length - 1)
        }
    }

    pub fn length(self, reader: &impl ReaderTrait) -> u64 {
        reader.read(field_ptr!(self, Queue<T>, length)).get()
    }
}

#[derive(Clone, Copy, FromBytes, FromZeroes, AsBytes, Unaligned)]
#[repr(packed)]
pub(crate) struct QueueElement<T: FromBytes + FromZeroes + AsBytes + Unaligned> {
    pub(crate) next: FilePointer<QueueElement<T>>,
    pub(crate) data: T,
    _phantom: PhantomData<T>,
}
