use std::{collections::HashSet, mem::size_of};

use zerocopy::{AsBytes, FromBytes, FromZeroes, Unaligned};

use crate::{
    mapped::Slabable, transaction, FilePointer, FileRange, RawFilePointer, ReaderTrait,
    TransactionHandle, U64,
};

#[derive(Clone, Copy, FromBytes, FromZeroes, AsBytes, Unaligned)]
#[repr(transparent)]
pub struct Vector {
    data: FileRange,
}

impl Slabable for Vector {
    fn slabs(slabs: &mut HashSet<u32>) {
        slabs.insert(size_of::<Vector>() as u32);
    }
}

impl Vector {
    pub fn new() -> Self {
        Self {
            data: RawFilePointer::null().range(0),
        }
    }
}

impl Default for Vector {
    fn default() -> Self {
        Self::new()
    }
}

impl Vector {
    pub fn set<R>(self, transaction: &mut TransactionHandle<R>, s: &[u8]) -> Self {
        let data = if s.is_empty() {
            Vector::new().data
        } else {
            let (range, data) = transaction.allocate_range(s.len() as u64);
            data.copy_from_slice(s.as_bytes());
            range
        };

        if self.data.len() != 0 {
            transaction.free_range(self.data);
        }

        Vector { data }
    }

    pub fn get(self, reader: &impl ReaderTrait) -> &[u8] {
        reader.read_raw(self.data)
    }
}

impl FilePointer<Vector> {
    pub fn set<R>(self, transaction: &mut TransactionHandle<R>, s: &[u8]) -> FilePointer<Vector> {
        let new_str = transaction.read::<Vector>(self).set(transaction, s);
        let (ptr, data) = transaction.modify(self);
        *data = new_str;
        ptr
    }

    pub fn get(self, reader: &impl ReaderTrait) -> &[u8] {
        reader.read(self).get(reader)
    }
}
